# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from typing import Union

import numpy as np

from parameterspace.priors.base import BasePrior
from parameterspace.utils import store_init_arguments


class Categorical(BasePrior):
    """Categorical prior with separate probabilities for each value."""

    @store_init_arguments
    def __init__(self, prior_values: Union[list, np.ndarray]):
        """
        Args:
            prior_values: Array containing the probabilities for each value.
        """
        super().__init__((0, 1))
        self.probabilities = np.array(prior_values)
        if not np.all(self.probabilities >= 0):
            raise ValueError("Probabilities must be non-negative!")

        self.probabilities = self.probabilities / np.sum(self.probabilities)
        self.numerical_values = (
            np.arange(len(self.probabilities), dtype=float) + 0.5
        ) / len(self.probabilities)

    def loglikelihood(self, value):
        return np.log(self.pdf(value))

    def pdf(self, value):
        """Probability given the numerical representation(s).

        The numerical representation is converted to integers and the corresponding
        prior_values are returned.
        If the value exceeds the given range of values, a ValueError is raised.

        Only finite values are converted, and the respective probability is computed.
        NaN and INF will result in a NaN. For example:
            value = [nan, inf, 0]
        will result in
            [NaN, NaN, p(0)].

        """
        value = np.atleast_1d(value)
        idx = np.isfinite(value)

        integer_value = np.around(len(self.probabilities) * value[idx] - 0.5).astype(
            int
        )
        pdf = np.full(value.shape, np.nan, dtype=float)

        try:
            pdf[idx] = self.probabilities[integer_value]
        except IndexError:
            raise ValueError(
                "Unknown value in the numerical representation for a "
                + "categorical parameter encountered!"
            )
        return pdf.squeeze()

    def sample(self, num_samples=None, random_state=np.random):
        return random_state.choice(
            self.numerical_values, size=num_samples, replace=True, p=self.probabilities
        )

    def __repr__(self):
        return "Categorical prior for %i values with p = %s" % (
            len(self.probabilities),
            self.probabilities,
        )

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and (len(self.probabilities) == len(other.probabilities))
            and np.allclose(self.probabilities, other.probabilities)
        )



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import abc
import importlib
from typing import Any, Optional, Tuple

import numpy as np


class BaseTransformation(abc.ABC):
    """Base class defining the API of a transformation."""

    def __init__(self, input_bounds: Optional[Tuple], output_bounds: Tuple):
        """
        Args:
            input_bounds: The numerical range for the input of a transformation. If
                that is not applicable, e.g. for categorical parameters, use `None`.
            output_bounds: The numerical range for the output of a transformation.
                As every transformation yields a numerical representation, these
                bounds have to be always specified in the form
                [lower_bound, upper_bound].
        """
        self.input_bounds = None if input_bounds is None else np.array(input_bounds)
        self.output_bounds = np.array(output_bounds)

    @abc.abstractmethod
    def inverse(self, numerical_value: float) -> Any:
        """Convert the numerical representation back to the true value with the proper
        type.

        Args:
            numerical_value: Transformed/Numerical representation of a value.

        Returns:
            The value corresponding to the given value. Type depends no the kind of
            transformation.
        """

    @abc.abstractmethod
    def __call__(self, value: Any) -> float:
        """Convert a value into the numerical representation.

        Args:
            value: A valid value for this transformation.

        Returns:
            Transformed/Numerical representation of the value.
        """

    def jacobian_factor(self, numerical_value: float) -> float:
        """Factor to correct the likelihood based on the non-linear transformation.

        Args:
            numerical_value: Transformed/Numerical representation of a value.

        Returns:
            Jacobian factor to properly transform the likelihood.
        """
        return 1.0

    @staticmethod
    def from_dict(json_dict: dict):
        """
        [summary]
        """
        transformation_class = json_dict["class_name"]
        module_str, class_str = transformation_class.rsplit(".", 1)
        module = importlib.import_module(module_str)
        model_class = getattr(module, class_str)
        return model_class(*json_dict["init_args"], **json_dict["init_kwargs"])

    def to_dict(self):
        """
        [summary]
        """
        json_dict = {
            "class_name": type(self).__module__ + "." + type(self).__qualname__,
            "init_args": self._init_args,
            "init_kwargs": self._init_kwargs,
        }
        return json_dict



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Union

import numpy as np

from parameterspace.parameters.base import BaseParameter
from parameterspace.priors.uniform import Uniform
from parameterspace.transformations.base import BaseTransformation
from parameterspace.transformations.categorical import Cat2Num
from parameterspace.utils import store_init_arguments


class OrdinalParameter(BaseParameter):
    """Ordinal parameter that can take discrete values of any type."""

    @store_init_arguments
    def __init__(
        self,
        name: str,
        values: list,
        *,
        prior: Union[list, np.ndarray, None] = None,
        transformation: Optional[BaseTransformation] = None,
        inactive_numerical_value: Optional[float] = np.nan
    ):
        """
        Initialize with options for ordinal parameter.

        Args:
            name: Name of the parameter.
            values: Allowed values in ascending order for this parameter.
            prior: List of (unnormalized) probablities for each value.
                Default puts equal probablity for each value.
            transformation: The only supported transformation right now
                is [parameterspace.transformations.categorical.Cat2Num][],
                which is also the default.
            inactive_numerical_value: Placeholder value for this parameter
                in case it is not active. Default is NaN.
        """
        self.values = values
        transformation = Cat2Num(values) if transformation is None else transformation
        prior = Uniform() if prior is None else prior
        super().__init__(
            name,
            prior,
            transformation,
            is_continuous=False,
            is_ordered=True,
            num_values=len(values),
            inactive_numerical_value=inactive_numerical_value,
        )

    def check_value(self, value):
        """Check if value is valid."""
        return value in self.values



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional, Tuple

import numpy as np

from parameterspace.transformations.base import BaseTransformation
from parameterspace.utils import store_init_arguments


class LogZeroOneFloat(BaseTransformation):
    """Maps bounded interval to [0,1] via a logarithmic transformation.

    This means that all the priors used with a parameter effectively model the
    exponent of the actual quantity.

    This class should be used for ContinuousParameters.
    """

    @store_init_arguments
    def __init__(self, bounds: Optional[Tuple]):
        """
        [summary]

        Args:
            bounds: Bounds must be positive for log-transformation.
        """
        assert min(bounds) > 0, "bounds must be positive for log-transformation"
        super().__init__(bounds, (0, 1))
        self.log_bounds = np.log(self.input_bounds)
        self.log_interval_size = self.log_bounds[1] - self.log_bounds[0]

    def inverse(self, numerical_value: float) -> float:
        return float(
            np.clip(
                np.exp(self.log_bounds[0] + numerical_value * (self.log_interval_size)),
                self.input_bounds[0],
                self.input_bounds[1],
            )
        )

    def __call__(self, value: Any) -> float:
        return float((np.log(value) - self.log_bounds[0]) / self.log_interval_size)

    def jacobian_factor(self, numerical_value: float) -> float:
        return 1.0 / (self.log_interval_size * self.inverse(numerical_value))

    def __eq__(self, other):
        return np.allclose(self.input_bounds, other.input_bounds)


class LogZeroOneInteger(BaseTransformation):
    """Maps a bounded interval of integers to [0, 1] via a logarithmic transformation.

    This means that all the priors used with a parameter effectively model the
    exponent of the actual quantity.

    This class should be used for IntegerParameters.
    """

    @store_init_arguments
    def __init__(self, bounds: Optional[Tuple]):
        """
        [summary]

        Args:
            bounds: Bounds must be positive for log-transformation.
        """
        assert min(bounds) > 0, "bounds must be positive for log-transformation"
        super().__init__(bounds, (0, 1))
        self.log_bounds = np.log(self.input_bounds + np.array([-0.5, 0.5]))
        self.log_interval_size = self.log_bounds[1] - self.log_bounds[0]

    def inverse(self, numerical_value: float) -> int:
        """
        [summary]
        """
        integer_value = np.around(
            np.exp(self.log_bounds[0] + numerical_value * (self.log_interval_size))
        )
        # clip result to bound due to rounding problems when the numerical value is 0.0
        return int(np.clip(integer_value, *self.input_bounds))

    def __call__(self, value: Any) -> float:
        return float((np.log(value) - self.log_bounds[0]) / self.log_interval_size)

    def jacobian_factor(self, numerical_value: float) -> float:
        return 1.0 / (self.log_interval_size * self.inverse(numerical_value))

    def __eq__(self, other):
        return np.allclose(self.input_bounds, other.input_bounds)



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

# For type hinting custom class (ParameterSpace):
from __future__ import annotations

import copy
import warnings
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np

import parameterspace as ps
from parameterspace.base import SearchSpace
from parameterspace.condition import Condition
from parameterspace.parameters.base import BaseParameter


class ParameterSpace(SearchSpace):
    """Class representing a parameter space that allows to sampling, converting and
    checking configurations.
    """

    def __init__(self, seed: int = None):
        super().__init__(seed=seed)

        self._parameters: dict = {}
        self._constants: dict = {}

    def __len__(self):
        return len(self._parameters)

    def __iter__(self):
        return iter(self._parameters.values())

    def __getitem__(self, name):
        return self._parameters[name]

    def __repr__(self):
        string = "ParameterSpace with the following parameteres and conditions\n"
        for p in self._parameters.values():
            string += str(p["parameter"])
            string += str(p["condition"]) + "\n"
        return string

    def __eq__(self, other):
        self_dict = self.to_dict()
        self_dict.pop("bit_generator_state")

        other_dict = other.to_dict()
        other_dict.pop("bit_generator_state")

        return self_dict == other_dict

    def seed(self, seed: Union[int, dict] = None) -> None:
        """Reinitialize the random number generator.

        Args:
            seed: Either an integer seed or a numpy bit generator state dictionary.
                Defaults to None.
        """
        if isinstance(seed, dict):
            self._rng = np.random.default_rng()
            self._rng.bit_generator.state = seed
        else:
            self._rng = np.random.default_rng(seed)

    def copy(self) -> ParameterSpace:
        return ParameterSpace.from_dict(self.to_dict())

    def add(
        self,
        parameter: Union[BaseParameter, ParameterSpace],
        condition: Callable = None,
    ):
        """Add a parameter that is only active if condition returns true when called.

        Args:
            parameter: A parameter or a subspace to be added to this space.\n
                Every parameter needs to have a unique name! A `ValueError` is raised if
                a parameter with the same name already exists in the `ParameterSpace`.
            condition: Lambda function that returns whether this parameter is active
                based on values of other parameters in the `ParameterSpace`.\n
                The signature is read using introspection to extract on which parameters
                the condition operates on.
        """
        if not isinstance(condition, Condition):
            tmp_condition = Condition(condition)
            condition = Condition()
            for n in tmp_condition.all_varnames:
                condition.merge(self._parameters[n]["condition"])

            condition.merge(tmp_condition)

        if isinstance(parameter, ParameterSpace):
            for pn in parameter.get_parameter_names():
                p = parameter[pn]
                tmp_condition = copy.deepcopy(condition).merge(p["condition"])
                self.add(p["parameter"], tmp_condition)
        else:
            if parameter.name in self._parameters:
                raise ValueError("Parameter %s already exists!" % parameter.name)
            self._parameters[parameter.name] = {
                "parameter": parameter,
                "condition": condition,
            }

    def to_dict(self) -> dict:
        """
        Transform current `ParameterSpace` into a dictionary representation.

        Returns:
            Contains "constants" and "parameters" (with "condition").
        """

        parameters = {}
        for p in self._parameters.values():
            parameters[p["parameter"].name] = {
                "parameter": p["parameter"].to_dict(),
                "condition": p["condition"].to_dict(),
            }
        return {
            "bit_generator_state": self._rng.bit_generator.state,
            "parameters": parameters,
            "constants": copy.deepcopy(self._constants),
        }

    @staticmethod
    def from_dict(dict_representation: dict) -> ParameterSpace:
        """
        Create a new `ParameterSpace` instance with the state passed as
        dictionary.

        Args:
            dict_representation: A dictionary representation of a
                `ParameterSpace`.

        Returns:
            New class instance
        """
        return_ps = ParameterSpace()
        return_ps.seed(dict_representation.get("bit_generator_state"))

        for parameter_dict in dict_representation["parameters"].values():
            parameter = BaseParameter.from_dict(parameter_dict["parameter"])
            condition = Condition.from_dict(parameter_dict["condition"])
            return_ps._parameters[parameter.name] = {
                "parameter": parameter,
                "condition": condition,
            }
        return_ps._constants = copy.deepcopy(dict_representation["constants"])

        return return_ps

    def fix(self, **kwargs: Any):
        """Remove all parameters from the parameters list and add treat them as
        constants.

        The main use-case for this method is to narrow down a generic
        `ParameterSpace` by fixing certain variables. This could be because of
        special constraints, or to focus any optimization on a few parameters.

        Note:
            This method will change the ParameterSpace object irreversibly.

        Args:
            **kwargs: All arguments to this method have to be keyword arguments
                with valid parameter names and valid values.\n
                The method will change the current `ParameterSpace` such that
                thenumerical representation does not contain any fixed
                parameters (or always inactive ones).\n
                The dictionary representation will still contain the fixed
                parameters with their respective value.

        Raises:
            ValueError: In case of not existing parameters or invalid values.

        """

        for name, value in kwargs.items():
            if name not in self._parameters:
                raise ValueError(
                    f"Parameter '{name}' is not part of this ParameterSpace!"
                )

            if not self._parameters[name]["parameter"].check_value(value):
                raise ValueError(f"Invalid value `{value}` for parameter '{name}'!")

        # remove any constants here that might not be active based on some parameter values,
        # but are still specified here.
        actual_constants = copy.copy(kwargs)
        for n in kwargs.keys():
            p = self._parameters[n]
            if p["condition"].all_varnames <= set(kwargs) and not p["condition"](
                kwargs
            ):
                actual_constants.pop(n, None)
        self._constants.update(**actual_constants)

        for p in list(self._parameters.keys()):

            if p in self._constants:
                del self._parameters[p]
                continue

            try:
                if self[p]["condition"](self._constants):
                    self[p]["condition"] = Condition(None)
                else:
                    del self._parameters[p]
            except ValueError:
                pass

    def remove(self, name: str):
        """Remove a parameter by name from this `ParameterSpace` instance.

        Note:
            When removing a parameter from a `ParameterSpace` that has been
            added to another `ParameterSpace`, the including `ParameterSpace`
            is not affected.

        Args:
            name: Name of the parameter.

        Raises:
            RuntimeError: In case other parameters condition on the parameter
                that should be removed.
            KeyError: In case parameter doesn't exist.
        """
        conditioning_parameters = []
        for p_name, p in self._parameters.items():
            if name in p["condition"].all_varnames:
                conditioning_parameters.append(p_name)
        if conditioning_parameters:
            raise RuntimeError(
                f"Unable to remove parameter '{name}' "
                + f"because the parameters {conditioning_parameters} condition on it. "
                + f"Please remove the other parameters before removing '{name}'."
            )

        try:
            self._parameters.pop(name)
        except KeyError as e:
            raise KeyError(
                f"Parameter '{name}' is not part of the ParameterSpace."
            ) from e

    def get_parameter_names(self) -> List[str]:
        """Get names of all parameters already in the current `ParameterSpace`.

        Returns:
            Parameter names.
        """
        return list(self._parameters.keys())

    def sample(
        self, partial_configuration: dict = None, rng: np.random.Generator = None
    ) -> dict:
        """Sample a random configuration based on the priors.

        Args:
            partial_configuration: Partial assignment of certain variables. If
                not `None`, only the remaining variables are set randomly. All
                conditions are honored.\n
                Right now, there is no check if the partitial configuration
                violates any conditions.
            rng: A Numpy random number generator used to generate the
                sample. Overrides the internal random number generator that is
                set on initialization of the space.

        Raises:
            ValueError: If a passed value is not valid for the corresponding
                parameter.

        Returns:
            Random parameter configuration.
        """
        if rng is None:
            rng = self._rng

        config = {} if partial_configuration is None else partial_configuration
        for n, v in config.items():
            if not self._parameters[n]["parameter"].check_value(v):
                raise ValueError("%s = %s is not valid for this space" % (n, v))

        config.update(self._constants)

        for n, p in self._parameters.items():
            if n in config:
                continue
            if p["condition"].empty() or p["condition"](config):
                config[n] = p["parameter"].sample_values(random_state=rng)

        return config

    def log_likelihood(self, configuration: dict) -> float:
        """Compute log-likelihood of a configuration under the given prior.

        Args:
            configuration: A parameter configuration

        Returns:
            Log-likelihood value or `NaN` if configuration is invalid.
        """
        if not self.check_validity(configuration):
            return np.nan
        self.remove_inactive(configuration)
        return self.log_likelihood_numerical(self.to_numerical(configuration))

    def log_likelihood_numerical(self, vector_configuration: np.ndarray) -> float:
        """Compute log-likelihood for the numerical representation of a
        configuration under the given prior.

        Note:
            This method assumes that the vector representation is valid! There
            are no sanity checks at this point. Inactive parameters must have
            `NaN` values.

        Args:
            vector_configuration: Numerical vector representation of a
                configuration.

        Returns:
            Log-likelihood value
        """
        ll = 0.0
        for i, p in enumerate(self._parameters.values()):
            if np.isfinite(vector_configuration[i]):
                ll += p["parameter"].loglikelihood_numerical_value(
                    vector_configuration[i]
                )
        return ll

    def check_validity(self, configuration: dict) -> bool:
        """Test whether the provided configuration is complete and valid.

        It checks wheter all parameters have a valid value or are not active.
        A warning is shown describing the issue.

        Args:
          configuration: The configuration to check.

        Returns:
            `True` if configuration is valid, `False` if not.
        """
        for n, p in self._parameters.items():
            v = configuration.get(n, None)
            if v is None:
                if p["condition"](configuration):
                    warnings.warn(
                        f"Parameter {n} should be active, "
                        + "but does not have a value assigned.\n"
                        + f"Full configuration: {configuration}",
                        RuntimeWarning,
                    )
                    return False
            else:
                if not p["condition"](configuration):
                    warnings.warn(
                        f"Parameter {n} = {v} should not be active!\n"
                        + f"Full configuration: {configuration}",
                        RuntimeWarning,
                    )
                    return False
                if not p["parameter"].check_value(v):
                    warnings.warn(
                        f"Parameter {n} = {v} is not a valid assignment!\n"
                        + f"Full configuration: {configuration}",
                        RuntimeWarning,
                    )
                    return False
        return True

    def remove_inactive(self, configuration: dict) -> dict:
        """Identify and remove parameters that should not have a value because
        their condition is unsatisfied.

        Args:
            configuration: Configuration to check.

        Returns:
            Cleaned configuration.
        """
        for n, p in self._parameters.items():
            if not p["condition"](configuration):
                configuration.pop(n, None)
        return configuration

    def to_numerical(self, configuration: dict) -> np.ndarray:
        self.remove_inactive(configuration)

        for name, value in self._constants.items():
            if name not in configuration:
                raise ValueError(
                    f"Configuration does not contain contant `{name} == {value}`!"
                )
            if configuration[name] != value:
                raise ValueError(
                    f"Constant parameter {name} has value {configuration[name]}, "
                    + f"but should be {value}!"
                )

        vec = np.zeros(len(self._parameters), dtype=float)
        for i, (n, p) in enumerate(self._parameters.items()):
            v = configuration.get(n, None)
            vec[i] = p["parameter"].val2num(v)
        return vec

    def val2num(self, configuration: dict) -> np.ndarray:
        """
        Attention:
            Deprecated. Use [ParameterSpace.to_numerical]\
            [parameterspace.parameterspace.ParameterSpace.to_numerical].

        Args:
            configuration: Configuration to convert.

        Raises:
            ValueError:  If configuration contains invalid names or values.

        Returns:
            Vector representation of the configuration.

        """
        warnings.warn(
            "Method ParameterSpace.val2num is deprecated and will be removed in "
            + "the future! Please use ParameterSpace.to_numerical instead.",
            category=DeprecationWarning,
        )
        return self.to_numerical(configuration)

    def from_numerical(self, vector: np.ndarray) -> dict:
        conf = {}
        for i, (n, p) in enumerate(self._parameters.items()):
            if not np.isnan(vector[i]):
                conf[n] = p["parameter"].num2val(vector[i])
        conf.update(self._constants)
        return self.remove_inactive(conf)

    def num2val(self, vector: np.ndarray) -> dict:
        """
        Attention:
            Deprecated. Use [ParameterSpace.from_numerical]\
            [parameterspace.parameterspace.ParameterSpace.from_numerical].

        Args:
            vector: Numerical vector representation of a configuration.

        Returns:
            Dictionary representation of the input configuration.

        """
        warnings.warn(
            "Method ParameterSpace.num2val is deprecated and will be removed in the "
            + "future! Please use ParameterSpace.from_numerical instead.",
            category=DeprecationWarning,
        )
        return self.from_numerical(vector)

    def has_conditions(self) -> bool:
        """Check if any of the parameters in the current `ParameterSpace` is
        conditioned on others.

        Returns:
            `True` of conditions exists, `False` if not.
        """
        for p in self:
            if not p["condition"].empty():
                return True
        return False

    def get_continuous_bounds(self) -> List[Tuple]:
        """Return the ParamerSpace specific bounds if it is purely continuous.

        Raises:
            ValueError: If `ParameterSpace` contains non-continuous parameters.

        Returns:
            Continuous bounds for this benchmark to easily run GP
        """
        bounds = []
        for n in self.get_parameter_names():
            p = self[n]["parameter"]
            if not p.is_continuous:
                raise ValueError(
                    "Parameterspace contains non-continuous parameter:\n{}".format(p)
                )
            bounds.append(tuple(p.get_numerical_bounds()))
        return bounds

    def to_latex_table(self, name_dict: Optional[dict] = None) -> str:
        """Construct LaTeX string for a table that shows the `ParameterSpace`
        with names, bounds and transformations.

        Args:
            name_dict: Mapping of parameter name in `ParameterSpace` to name
                to be rendered in LaTeX table.

        Returns:
            LaTeX table represenation
        """
        try:
            from num2tex import configure as num2tex_configure
            from num2tex import num2tex
        except ImportError as e:
            raise RuntimeError(
                "To use this functionality, please install num2tex."
            ) from e

        num2tex_configure(exp_format="cdot")

        name_dict = {} if name_dict is None else name_dict

        latex_strs = [
            "\\begin{tabular}{c c c c c}",
            "\\hline",
            "Parameter Name & Type & Values & Transformation & Prior \\\\",
            "\\hline",
        ]

        for parameter_name in self._parameters:
            parameter = self._parameters[parameter_name]["parameter"]

            name_str = name_dict.get(parameter_name, parameter.name)
            transformation_name = type(parameter._transformation).__name__
            prior_name = type(parameter._prior).__name__

            if isinstance(parameter, ps.IntegerParameter):
                type_str = "Integer"
                values_str = "$[{0[0]}, {0[1]}]$".format(parameter.bounds)
                transformation_str = "Log" if "Log" in transformation_name else ""
                prior_str = prior_name

            if isinstance(parameter, ps.ContinuousParameter):
                type_str = "Float"
                prior_str = prior_name

                if "Log" in transformation_name:
                    transformation_str = "Log"
                    values_str = "$[{}, {}]$".format(
                        num2tex(parameter.bounds[0], precision=2),
                        num2tex(parameter.bounds[1], precision=2),
                    )
                else:
                    transformation_str = " "
                    values_str = "$[{0[0]}, {0[1]}]$".format(parameter.bounds)

            if isinstance(parameter, ps.CategoricalParameter):
                type_str = "Categorical"
                values_str = "[" + ", ".join(parameter.values) + "]"
                transformation_str = " "
                prior_probs = parameter._prior.probabilities
                prior_str = (
                    "["
                    + ",".join(map(lambda p: "{:3.2f}".format(p), prior_probs))
                    + "]"
                )

            latex_strs.append(
                " & ".join(
                    [name_str, type_str, values_str, transformation_str, prior_str]
                )
                + "\\\\"
            )

        latex_strs.extend(["\\hline", "\\end{tabular}"])
        return "\n".join(latex_strs)



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from parameterspace.priors.beta import Beta
from parameterspace.priors.categorical import Categorical
from parameterspace.priors.truncated_normal import TruncatedNormal
from parameterspace.priors.uniform import Uniform



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from parameterspace.transformations.categorical import Cat2Num
from parameterspace.transformations.log_zero_one import (
    LogZeroOneFloat,
    LogZeroOneInteger,
)
from parameterspace.transformations.zero_one import ZeroOneFloat, ZeroOneInteger



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from parameterspace.parameters.categorical import CategoricalParameter
from parameterspace.parameters.continuous import ContinuousParameter
from parameterspace.parameters.integer import IntegerParameter
from parameterspace.parameters.ordinal import OrdinalParameter

__all__ = [
    "ContinuousParameter",
    "CategoricalParameter",
    "IntegerParameter",
    "OrdinalParameter",
]



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional, Tuple, Union

import numpy as np

from parameterspace.parameters.base import BaseParameter
from parameterspace.priors.categorical import Categorical
from parameterspace.transformations.base import BaseTransformation
from parameterspace.transformations.categorical import Cat2Num
from parameterspace.utils import store_init_arguments


class CategoricalParameter(BaseParameter):
    """Categorical parameter that can take discrete values of any type."""

    @store_init_arguments
    def __init__(
        self,
        name: str,
        values: Tuple[Any, ...],
        prior: Union[list, np.ndarray, None] = None,
        transformation: Optional[BaseTransformation] = None,
        inactive_numerical_value: Optional[float] = np.nan,
    ):
        """
        Initialize with options for categorical parameter.

        Args:
            name: Name of the parameter.
            values: Allowed values for this parameter.
            prior: Probabilities for each value (does not need to be normalized).
            transformation: A transformation that can translate the arbitrary type of
                the values to a numerical value. The only supported one right now
                is [parameterspace.transformations.categorical.Cat2Num][], which is
                also used as default.
            inactive_numerical_value: [description]
        """
        self.values = values
        transformation = Cat2Num(values) if transformation is None else transformation

        if prior is None:
            prior = Categorical([1.0] * len(values))
        elif not isinstance(prior, Categorical):
            prior = Categorical(prior)

        super().__init__(
            name,
            prior,
            transformation,
            is_continuous=False,
            is_ordered=False,
            num_values=len(values),
            inactive_numerical_value=inactive_numerical_value,
        )

    def check_value(self, value):
        """Check if value is valid."""
        return value in self.values



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import scipy.stats as sps

from parameterspace.priors.base import BasePrior


class Normal(BasePrior):
    """Stardand Gaussian prior."""

    def __init__(self, mean=0, std=1):
        super().__init__(np.array([-np.inf, np.inf]))
        self.sps_normal_dist = sps.norm(loc=mean, scale=std)

    def loglikelihood(self, value):
        return self.sps_normal_dist.logpdf(value)

    def pdf(self, value):
        return self.sps_normal_dist.pdf(value)

    def sample(self, num_samples=None, random_state=np.random):
        return self.sps_normal_dist.rvs(size=num_samples, random_state=random_state)

    def __eq__(self, other):
        return super().__eq__(other) and np.allclose(
            [self.mean, self.std], [other.mean, other.std]
        )



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import abc
import copy
import importlib
from typing import Optional, Union

import numpy as np

from parameterspace.priors.base import BasePrior
from parameterspace.transformations.base import BaseTransformation


class BaseParameter(abc.ABC):
    """Base parameter class defining the API."""

    def __init__(
        self,
        name: str,
        prior: BasePrior,
        transformation: BaseTransformation,
        is_continuous: bool,
        is_ordered: bool,
        num_values: Union[int, float],
        inactive_numerical_value: Optional[float] = np.nan,
    ):
        """
        Initialize with options common to all parameters.

        Args:
            name: Name of the parameter.
            prior: Defines the pdf from which random samples are drawn. Default (`None`)
                correponds to a uniform prior. Note that the prior's bounds must match
                the transformation's output bounds.
            transformation: Defines the tranformation from the values to their numerical
                representaiton. Note that the transformation's output bounds must match
                the prior's bounds.
            is_continuous: Indicates whether this parameter varies continuously, i.e.
                is a `float`.
            is_ordered: Indicates whether this parameter has a natural ordering of it's
                values.
            num_values: Number of possible values. For ordinal, categorical, and integer
                parameters, this equals the number of unique values. For continuous
                parameters, it equals np.inf (eventhough technically there is only a
                finite number of float values).
            inactive_numerical_value: Placeholder value for this parameter in case it
                is not active
        """
        self.name = name
        self._prior = prior
        self._transformation = transformation
        self._inactive_numerical_value = inactive_numerical_value
        self.transformed_bounds = self._transformation.output_bounds

        assert np.allclose(
            self._prior.bounds, self.transformed_bounds, 1e-6
        ), "Missmatch between the prior bounds (%s)" % (
            self._prior.bounds
        ) + "and the transformation's bounds (%s)!" % (
            self.transformed_bounds
        )

        self.is_continuous = is_continuous
        self.is_ordered = is_ordered
        self.num_values = num_values

    def __repr__(self):
        """Basic info about a parameter common to all types."""
        string = "Name: %s\n" % self.name
        string += "Type: %s\n" % self.__class__
        string += "Prior: " + self._prior.__repr__() + "\n"
        string += "is continuous: %s\n" % self.is_continuous
        string += "is ordered: %s\n" % self.is_ordered
        string += "num_values: %s\n" % self.num_values
        return string

    def sample_values(self, num_samples=None, random_state=np.random):
        """Generate randomly sampled values based on the prior."""
        numerical_samples = self.sample_numerical_values(num_samples, random_state)
        if num_samples is None:
            return self._transformation.inverse(numerical_samples)
        return [
            self._transformation.inverse(num_value) for num_value in numerical_samples
        ]

    def sample_numerical_values(self, num_samples=None, random_state=np.random):
        """Generate random values based on the prior, but in the transformed space."""
        return self._prior.sample(num_samples, random_state=random_state)

    def val2num(self, value):
        """Translate a value into its numerical representation (incl. normalization)."""
        if value is None:
            return self._inactive_numerical_value
        return self._transformation(value)

    def num2val(self, numerical_value):
        """Translate the numerical representation into the actual value."""
        return self._transformation.inverse(numerical_value)

    def check_numerical_value(self, numerical_value):
        """Check if the numerical representation of the value is valid."""
        return (
            self.transformed_bounds[0] <= numerical_value <= self.transformed_bounds[1]
        )

    def pdf_numerical_value(self, numerical_value):
        """Compute the PDF based on the prior."""
        return self._prior.pdf(numerical_value) * self._transformation.jacobian_factor(
            numerical_value
        )

    def loglikelihood_numerical_value(self, numerical_value):
        """Compute the loglikelihood based on the prior."""
        return self._prior.loglikelihood(numerical_value) + np.log(
            self._transformation.jacobian_factor(numerical_value)
        )

    def loglikelihood(self, value):
        """Compute the loglikelihood based on the prior."""
        return self.loglikelihood_numerical_value(self._transformation(value))

    def get_numerical_bounds(self):
        """Translate the provided bounds into the numerical representation."""
        return self.transformed_bounds

    @abc.abstractmethod
    def check_value(self, value):
        """Checks if value is valid."""

    @staticmethod
    def from_dict(json_dict):
        parameter_class = json_dict["class_name"]
        module_str, class_str = parameter_class.rsplit(".", 1)
        module = importlib.import_module(module_str)
        parameter_class = getattr(module, class_str)

        transformation = BaseTransformation.from_dict(json_dict["transformation"])
        prior = BasePrior.from_dict(json_dict["prior"])

        return parameter_class(
            *json_dict["init_args"],
            **json_dict["init_kwargs"],
            transformation=transformation,
            prior=prior
        )

    def to_dict(self):
        json_dict = {
            "class_name": type(self).__module__ + "." + type(self).__qualname__,
            "init_args": self._init_args,
            "init_kwargs": copy.deepcopy(self._init_kwargs),
            "transformation": self._transformation.to_dict(),
            "prior": self._prior.to_dict(),
        }
        for key in ["transformation", "prior"]:
            if key in json_dict["init_kwargs"]:
                del json_dict["init_kwargs"][key]

        return json_dict

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.name != other.name:
            return False
        return (self._prior == other._prior) and (
            self._transformation == other._transformation
        )



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from typing import Optional, Tuple, Union

import numpy as np

from parameterspace.parameters.base import BaseParameter
from parameterspace.priors.base import BasePrior
from parameterspace.priors.uniform import Uniform
from parameterspace.transformations.base import BaseTransformation
from parameterspace.transformations.log_zero_one import LogZeroOneInteger
from parameterspace.transformations.zero_one import ZeroOneInteger
from parameterspace.utils import store_init_arguments


class IntegerParameter(BaseParameter):
    """Integer parameter that can take every integer value in a given range."""

    @store_init_arguments
    def __init__(
        self,
        name: str,
        bounds: Tuple,
        *,
        prior: Optional[BasePrior] = None,
        transformation: Union[BaseTransformation, str, None] = None,
        inactive_numerical_value: Optional[float] = np.nan
    ):
        """
        Initialize with options for integer parameter.

        Args:
            name: Name of the parameter.
            bounds: Lower and upper bound for the parameter.
            prior: Default prior is the UniformInteger prior.
            transformation: Default transformation is the identity. If a string is
                given, the transform is created automatically for supported transforms.
            inactive_numerical_value: Placeholder value for this parameter in case it
                is not active.
        """
        self.bounds = np.array(bounds, dtype=int)

        if not isinstance(transformation, BaseTransformation):
            if transformation is None:
                transformation = ZeroOneInteger(self.bounds)
            elif transformation == "log":
                transformation = LogZeroOneInteger(self.bounds)

        prior = Uniform() if prior is None else prior

        # TODO: Ask Stefan: inactive_numerical_value is unused/not passed
        super().__init__(
            name,
            prior,
            transformation,
            is_continuous=False,
            is_ordered=True,
            num_values=bounds[1] - bounds[0] + 1,
            inactive_numerical_value=np.nan,
        )

    def __repr__(self):
        """Add bounds to the string representation."""
        string = super().__repr__()
        string += "Bounds: [{}, {}]\n".format(self.bounds[0], self.bounds[1])
        return string

    def check_value(self, value):
        """Check if value is valid."""
        return self.bounds[0] <= value <= self.bounds[1] and value == int(value)

    def num2val(self, numerical_value):
        """Translate the numerical representation into the actual value."""
        return int(super().num2val(numerical_value))



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional, Tuple

import numpy as np

from parameterspace.transformations.base import BaseTransformation
from parameterspace.utils import store_init_arguments


class ZeroOneFloat(BaseTransformation):
    """Maps a bounded interval to [0, 1] via a linear transformation."""

    @store_init_arguments
    def __init__(self, bounds: Optional[Tuple]):
        super().__init__(bounds, (0, 1))
        self.interval_size = bounds[1] - bounds[0]

    def inverse(self, numerical_value: float) -> float:
        return float(
            np.clip(
                self.input_bounds[0] + numerical_value * (self.interval_size),
                self.input_bounds[0],
                self.input_bounds[1],
            )
        )

    def __call__(self, value: Any) -> float:
        return float((value - self.input_bounds[0]) / self.interval_size)

    def __eq__(self, other):
        return np.allclose(self.input_bounds, other.input_bounds)

    def jacobian_factor(self, numerical_value: float) -> float:
        return 1.0 / self.interval_size


class ZeroOneInteger(BaseTransformation):
    """Maps a bounded interval of integers to [0, 1] via a linear transformation."""

    @store_init_arguments
    def __init__(self, bounds: Optional[Tuple]):
        super().__init__(bounds, (0, 1))
        self.interval_size = bounds[1] - bounds[0] + 1

    def inverse(self, numerical_value: float) -> int:
        return int(
            np.clip(
                np.around(
                    self.input_bounds[0] - 0.5 + numerical_value * (self.interval_size)
                ),
                self.input_bounds[0],
                self.input_bounds[1],
            )
        )

    def __call__(self, value: Any) -> float:
        return float((value - self.input_bounds[0] + 0.5) / self.interval_size)

    def __eq__(self, other):
        return np.allclose(self.input_bounds, other.input_bounds)

    def jacobian_factor(self, numerical_value: float) -> float:
        return 1.0 / self.interval_size



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import numpy.random as npr
import scipy.stats as sps

from parameterspace.priors.base import BasePrior
from parameterspace.utils import store_init_arguments


class TruncatedNormal(BasePrior):
    """Truncated normal prior for bounded parameters."""

    @store_init_arguments
    def __init__(self, mean: float = 0.5, std: float = 1):
        """
        Args:
            mean: Mean of the distribution.
            std: Standard deviation of the distribution
        """
        super().__init__((0, 1))
        a, b = (self.bounds - mean) / std
        self.mean = mean
        self.std = std
        self.sps_dist = sps.truncnorm(a, b, loc=mean, scale=std)

    def pdf(self, value):
        return self.sps_dist.pdf(value)

    def loglikelihood(self, value):
        return self.sps_dist.logpdf(value)

    def sample(self, num_samples=None, random_state=npr):
        return self.sps_dist.rvs(size=num_samples, random_state=random_state)

    def __repr__(self):
        return (
            "Truncated normal (Interval [%f, %f]) with parameteres mean=%f, std=%f"
            % (self.bounds[0], self.bounds[1], self.mean, self.std)
        )

    def __eq__(self, other):
        return super().__eq__(other) and np.allclose(
            [self.mean, self.std], [other.mean, other.std]
        )



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import scipy.stats as sps

from parameterspace.priors.base import BasePrior
from parameterspace.utils import store_init_arguments


class Beta(BasePrior):
    """Beta prior for variables in the interval [0,1]."""

    @store_init_arguments
    def __init__(self, a: float, b: float):
        """
        Args:
            a: Positive parameter of the Beta distribution.
            b: Positive parameter of the Beta distribution.
        """
        super().__init__((0, 1))
        self.a, self.b = a, b
        self.sps_beta_dist = sps.beta(a, b)

    def pdf(self, value):
        return self.sps_beta_dist.pdf(value)

    def loglikelihood(self, value):
        return self.sps_beta_dist.logpdf(value)

    def sample(self, num_samples=None, random_state=np.random):
        return self.sps_beta_dist.rvs(size=num_samples, random_state=random_state)

    def __repr__(self):
        return "Beta distribution with parameteres a=%f, b=%f" % (self.a, self.b)

    def __eq__(self, other):
        return super().__eq__(other) and np.allclose(
            [self.a, self.b], [other.a, other.b]
        )



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import abc
import importlib
from typing import Tuple

import numpy as np


class BasePrior(abc.ABC):
    """Base class defining the API of the priors.

    The priors enable the incorporation of domain knowledge into the parameter
    definition by allowing the specification of a PDF/PMF. These are used to sample
    random values and to compute the loglikelihood of a given value.
    """

    def __init__(self, bounds: Tuple):
        """
        Args:
            bounds: Lower and upper bound of the prior.
        """
        self.bounds = np.array(bounds, dtype=float)

    def loglikelihood(self, value):
        """
        Compute the log PDF (up to an additive constant) of a given value.

        Note:
            Values for the priors are always after the transformation!

        Args:
            value: [description]

        Returns:
            [descriptions]
        """
        return np.log(self.pdf(value))

    @abc.abstractmethod
    def pdf(self, value):
        """
        Computes the PDF of a given value.

        Note:
            Values for the priors are always after the transformation!

        Args:
            value: [description]

        Returns:
            [descriptions]
        """

    @abc.abstractmethod
    def sample(self, num_samples: int = 1):
        """
        Draw random samples from the prior.

        Args:
            num_samples: [description]

        Returns:
            [descriptions]
        """

    @staticmethod
    def from_dict(json_dict):
        prior_class = json_dict["class_name"]
        module_str, class_str = prior_class.rsplit(".", 1)
        module = importlib.import_module(module_str)
        model_class = getattr(module, class_str)
        return model_class(*json_dict["init_args"], **json_dict["init_kwargs"])

    def to_dict(self):
        json_dict = {
            "class_name": type(self).__module__ + "." + type(self).__qualname__,
            "init_args": self._init_args,
            "init_kwargs": self._init_kwargs,
        }
        return json_dict

    def __eq__(self, other):
        """Uniform prior doesn't have a state, so equality is just class membership."""
        return isinstance(other, type(self))



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from parameterspace.priors.base import BasePrior
from parameterspace.utils import store_init_arguments


class Uniform(BasePrior):
    """Uninformed prior that puts equal weight on every value."""

    @store_init_arguments
    def __init__(self):
        super().__init__([0, 1])

    def pdf(self, value):
        """Calculate probability density function value.

        Return constant for values inside the bounds, zero if outside, and NaN for NaNs.
        """
        value = np.atleast_1d(value)
        active_idx = np.isfinite(value)
        pdf = np.full(value.shape, np.nan)
        inside = np.logical_and(
            self.bounds[0] <= value[active_idx], value[active_idx] <= self.bounds[1]
        )
        pdf[active_idx] = 1.0 / (self.bounds[1] - self.bounds[0]) * (inside)
        return pdf.squeeze()

    def sample(self, num_samples=None, random_state=np.random):
        return random_state.uniform(
            low=self.bounds[0], high=self.bounds[1], size=num_samples
        )

    def __repr__(self):
        """Minimal information about the Prior."""
        return "Uniform prior in the interval [%f, %f]." % (
            self.bounds[0],
            self.bounds[1],
        )



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import parameterspace.priors
import parameterspace.transformations
from parameterspace.parameters import *
from parameterspace.parameterspace import *



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import inspect
from typing import Callable

from parameterspace import utils


class Condition:
    """Class that handels the logic of conditioning parameters on others in a
    ParameterSpace."""

    def __init__(self, lambda_fn: Callable = None):
        """
        Args:
            lambda_fn: A callable whose signature contains valid names of other
                parameters. It returns True if the associate parameter is active,
                and False otherwise.
        """
        self.functions = []
        self.function_texts = []
        self.varnames = []
        self.all_varnames = set()

        if lambda_fn is not None:
            self.functions = [lambda_fn]
            varnames, function_text = utils.extract_lambda_information(
                inspect.getsourcelines(lambda_fn)[0]
            )
            # self.varnames = [lambda_fn.__code__.co_varnames]
            self.varnames = [varnames]
            self.function_texts = [function_text]
            self.all_varnames = set(lambda_fn.__code__.co_varnames)

    def __repr__(self):
        if not self.all_varnames:
            return "No conditions!"
        string = "Condition(s) depend(s) on {}.\n".format(
            self.all_varnames
        ) + "\n".join(self.function_texts)
        return string

    def __call__(self, config=None):
        if not self.all_varnames:
            return True
        config = {} if config is None else config
        for (vn, fn) in zip(self.varnames, self.functions):
            if not set(vn).issubset(config.keys()):
                raise ValueError("Not all variables set to evaluate the condition!")
            args = [config[n] for n in vn]
            if None in args:
                return False
            if not fn(*args):
                return False
        return True

    def empty(self) -> bool:
        """Check if this Condition is not trivial.

        Returns:
            [description]
        """
        return not bool(self.all_varnames)

    def merge(self, other: Condition = None) -> Condition:
        """Concatenates two Conditions to allow for hierarchical spaces.

        Args:
            other: [description]

        Returns:
            [description]
        """
        if other is not None:
            self.all_varnames = self.all_varnames.union(other.all_varnames)
            self.functions.extend(other.functions)
            self.function_texts.extend(other.function_texts)
            self.varnames.extend(other.varnames)
        return self

    def to_dict(self) -> dict:
        """[summary]

        Returns:
            [description]
        """
        return_dict = {}
        for i, (varnames, function_text) in enumerate(
            zip(self.varnames, self.function_texts)
        ):
            return_dict[str(i)] = {"varnames": varnames, "function_text": function_text}
        return return_dict

    @staticmethod
    def from_dict(
        dict_representation: dict, verify_lambda: Callable = utils.verify_lambda
    ) -> Condition:
        """[summary]

        Args:
            dict_representation: [description]
            verify_lambda: [description]

        Raises:
            RuntimeError: If "function_text" in in `dict_representation` is
                not considered save.

        Returns:
            [description]
        """

        return_condition = Condition()

        for k in range(len(dict_representation)):
            varnames = dict_representation[str(k)]["varnames"]
            function_text = dict_representation[str(k)]["function_text"]

            if not verify_lambda(varnames, function_text):
                raise RuntimeError(f"Source {function_text} is not considered save!")

            return_condition.function_texts.append(function_text)
            return_condition.varnames.append(varnames)
            return_condition.all_varnames |= set(varnames)

            # pylint: disable=eval-used
            return_condition.functions.append(
                eval(f"lambda {', '.join(varnames)}: {function_text}")
            )
            # pylint: enable=eval-used

        return return_condition

    def __eq__(self, other):
        if self.all_varnames != other.all_varnames:
            return False
        if len(self.function_texts) != len(other.function_texts):
            return False

        for t1, t2 in zip(self.function_texts, other.function_texts):
            if t1 != t2:
                return False

        return True



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import copy
import math
from typing import Any

import numpy as np

from parameterspace.transformations.base import BaseTransformation
from parameterspace.utils import store_init_arguments


class Cat2Num(BaseTransformation):
    """Translates any values into discrete, equidistant values between 0 and 1."""

    @store_init_arguments
    def __init__(self, values: list):
        """
        Args:
            values: List of all possible values of (almost) arbitrary type.
        """
        super().__init__(None, (0, 1))
        self.values = copy.deepcopy(values)

    def inverse(self, numerical_value: float) -> Any:
        return self.values[
            int(
                np.clip(
                    np.around(numerical_value * len(self.values) - 0.5),
                    0,
                    len(self.values) - 1,
                )
            )
        ]

    def __call__(self, value: Any) -> float:
        idx = self.values.index(value)
        return (0.5 + idx) / len(self.values)

    def __eq__(self, other):
        if len(self.values) != len(other.values):
            return False
        for e1, e2 in zip(self.values, other.values):
            # need to handle some special cases that can occur with categorical values
            if (e1 is None) and (e2 is None):
                continue
            if e1 == e2:
                continue
            # == will not yield true, if both values are NAN
            try:
                if math.isnan(e1) and math.isnan(e2):
                    continue
            except TypeError:
                pass

            return False

        return True



import abc

import numpy as np


class SearchSpace(abc.ABC):
    """Abstract base class for different search space implementations."""

    def __init__(self, seed: int = None) -> None:
        """Initialize a search space with an optional seed."""
        self.seed(seed)

    @abc.abstractmethod
    def __len__(self) -> int:
        """The number of parameters in the space."""

    @abc.abstractmethod
    def seed(self, seed: int) -> None:
        """Reinitialize the random number generator with a new seed."""

    @abc.abstractmethod
    def copy(self):
        """Get a copy that behaves exactly like the original `SearchSpace`.
        Call `seed()` on the copy to get independent samples.
        """

    @abc.abstractmethod
    def sample(self) -> dict:
        """Provide a dictionary with one key corresponding to each parameter name and
        its value representing a sample for that parameter.
        """

    @abc.abstractmethod
    def to_numerical(self, configuration: dict) -> np.ndarray:
        """Given a configuration from this space, create a numerical vector
        representation.
        The transformed representation needs to be between 0 and 1 (uniform), including
        integers, ordinal and categoricals.
        Inactive parameters have to be represented with `np.nan`
        """

    @abc.abstractmethod
    def from_numerical(self, vector: np.ndarray) -> dict:
        """Convert a np.float64 type vector numerical representation of a configuration
        from this space to a dictionary representation."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parameterspace',
 'parameterspace.parameters',
 'parameterspace.priors',
 'parameterspace.transformations']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.17.0', 'scipy>=1.6.0']

setup_kwargs = {
    'name': 'parameterspace',
    'version': '0.7.14',
    'description': 'Parametrized hierarchical spaces with flexible priors and transformations.',
    'long_description': '# ParameterSpace\n\n**Contents:**\n\n- [About](#about)\n- [Installation](#installation)\n- [Development](#development)\n  - [Prerequisites](#prerequisites)\n  - [Setup environment](#setup-environment)\n  - [Running Tests](#running-tests)\n  - [Building Documentation](#building-documentation)\n- [License](#license)\n\n## About\n\nA package to define parameter spaces consisting of mixed types (continuous, integer,\ncategorical) with conditions and priors. It allows for easy specification of the\nparameters and their dependencies. The `ParameterSpace` object can then be used to\nsample random random configurations from the prior and convert any valid configuration\ninto a numerical representation. This numerical representation has the following\nproperties:\n\n- it results in a Numpy `ndarray` of type `numpy.float64`\n- transformed representation between 0 and 1 (uniform) including integers, ordinal and\n  categorical parameters\n- inactive parameters are masked as `numpy.nan` values\n\nThis allows to easily use optimizers that expect continuous domains to be used on more\ncomplicated problems because `parameterspace` can convert any numerical vector\nrepresentation inside the unit hypercube into a valid configuration. The function might\nnot be smooth, but for robust methods (like genetic algorithms/evolutionary strategies)\nthis might still be valuable.\n\nThis software is a research prototype. The software is not ready for production use. It\nhas neither been developed nor tested for a specific use case. However, the license\nconditions of the applicable Open Source licenses allow you to adapt the software to\nyour needs. Before using it in a safety relevant setting, make sure that the software\nfulfills your requirements and adjust it according to any applicable safety standards\n(e.g. ISO 26262).\n\n## Documentation\n\n**Visit [boschresearch.github.io/parameterspace](https://boschresearch.github.io/parameterspace/)**\n\n\n## Installation\n\nThe `parameterspace` package can be installed from [pypi.org](https://pypi.org):\n\n```\npip install parameterspace\n```\n\n## Development\n\n### Prerequisites\n\n- Python >= 3.7\n- [Poetry](https://python-poetry.org/docs/#installation)\n\n### Setup environment\n\nTo install the package and its dependencies for development run:\n\n```\npoetry install\n```\n\nOptionally install [pre-commit](https://pre-commit.com) hooks to check code standards\nbefore committing changes:\n\n```\npoetry run pre-commit install\n```\n\n### Running Tests\n\nThe tests are located in the `./tests` folder. The [pytest](https://pytest.org)\nframework is used for running them. To run the tests:\n\n```\npoetry run pytest ./tests\n```\n\n### Building Documentation\n\nTo built documentation run from the repository root:\n\n```\npoetry run mkdocs build --clean\n```\n\nFor serving it locally while working on the documentation run:\n\n```\npoetry run mkdocs serve\n```\n\n## License\n\n`parameterspace` is open-sourced under the Apache-2.0 license. See the\n[LICENSE](LICENSE) file for details.\n\nFor a list of other open source components included in `parameterspace`, see the file\n[3rd-party-licenses.txt](3rd-party-licenses.txt).\n',
    'author': 'Bosch Center for AI, Robert Bosch GmbH',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/boschresearch/parameterspace',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright (c) 2021 - for information on the respective copyright owner see the
# NOTICE file and/or the repository https://github.com/boschresearch/parameterspace
#
# SPDX-License-Identifier: Apache-2.0

import os
import re
from functools import wraps
from typing import Callable, Iterable, Tuple

import numpy as np


def store_init_arguments(init_method: Callable) -> Callable:
    """Stores init arguments (including keyword arguments) and auto converts
    `numpy.ndarray`s to `list`s for json serializability.

    Args:
        init_method: [description]

    Returns:
        [description]
    """

    @wraps(init_method)
    def wrapper(self, *args, **kwargs):
        args = [a.tolist() if isinstance(a, np.ndarray) else a for a in args]

        for k, k_value in kwargs.items():
            if isinstance(k_value, np.ndarray):
                kwargs[k] = k_value.tolist()

        self._init_args = args
        self._init_kwargs = kwargs
        init_method(self, *args, **kwargs)

    return wrapper


def extract_lambda_information(source_lines: Iterable) -> Tuple[list, str]:
    """[summary]

    Args:
        source_lines: [description]

    Raises:
        RuntimeError: If function definition is not a valid Lambda function.

    Returns:
        [description]
    """
    condensed_code = "".join(source_lines).replace(os.linesep, "")  # join lines
    condensed_code = " ".join(
        condensed_code.split()
    )  # replace multiple spaces by a single one

    try:
        signature, body = condensed_code.split("lambda")[1].split(":")
    except IndexError as e:
        raise RuntimeError(
            "The function definition does not look like a valid Lambda function:\n"
            + "".join(source_lines)
        ) from e

    while len(body) > 1:
        lambda_def = f"lambda {signature}: {body}"
        try:
            _ = compile(lambda_def, "<unused filename>", "eval")
            break
        except SyntaxError:
            body = body[:-1]

    variables = [s.strip() for s in signature.split(",")]

    return variables, body.strip().rstrip(",")


def verify_lambda(variables: list, body: str) -> bool:
    """Check serialized lambda expression for malicious code.

    Args:
        variables: [description]
        body: [description]

    Returns:
        [description]
    """

    if len(body) > 200:
        return False

    if "eval(" in body:
        return False

    blacklisted_characters = "\\;"

    for c in blacklisted_characters:
        if c in body:
            return False

    white_listed_chars = ".0123456789+-*/()<=!> "
    white_listed_functions = [
        "math.sin",
        "math.cos",
        "math.exp",
        "math.log",
        "or",
        "and",
        "not",
    ]

    for vn in variables:
        body = body.replace(vn, "")

    for fn in white_listed_functions:
        body = body.replace(fn, "")

    for c in white_listed_chars:
        body = body.replace(c, "")

    # remove all single quoted strings
    matches = re.findall(r"\'(.+?)\'", body)
    for m in matches:
        body = body.replace("'" + m + "'", "")

    # remove all double quoted strings
    matches = re.findall(r"\"(.+?)\"", body)
    for m in matches:
        body = body.replace('"' + m + '"', "")

    if len(body) > 0:
        print("body:", body)
        return False

    return True
